/*
btisrt.c Version 0.1.0  Btree insert record
Copyright (C) 2003  dondalah@ripco.com (Dondalah)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to:

	Free Software Foundation, Inc.
	59 Temple Place - Suite 330
	Boston, MA  02111-1307, USA.
*/

/* return code 0 is good insert */
/* return code 1 is duplicate record */

#include <bthash.h>

int btisrt(btreefmt *bthdr, long ofst,
   unsigned char *rcd)
   {
   int currcd;
   int sigrslt;
   int rslt;
   unsigned char *p,*q,*r;
   unsigned char *rcdptr;
   bt_nodefmt *node;
   bthdr->reason = 0;
   sigrslt = strcmp(bthdr->signature,BT_SIGNATURE);
   if (sigrslt)
      {
      fprintf(stderr,"btisrt: non-btree "
         "insertion\n");
      bthdr->reason = BTREE_ERR + 1022;
      return(BTREE_ERR + 1022);
      } /* not btree database */
   p = (unsigned char *) rcd + bthdr->dletofst;
   if (*p == DELETED)
      {
      fprintf(stderr,"btisrt: delete byte set\n");
      bthdr->reason = BTREE_ERR + 1019;
      return(BTREE_ERR + 1019);
      } /* delete byte set */
   if (ofst == ROOT
      && bthdr->head->next != bthdr->tail)
      {
      fprintf(stderr,"btisrt: inserting at root "
	 "with non-empty stack\n");
      bthdr->reason = BTREE_ERR + 1001;
      return(BTREE_ERR + 1001);
      } /* insert loop */
   btgetblk(bthdr,ofst);
   if (bthdr->reason)
      return(BTREE_ERR + 1002);
   btpush(bthdr,ofst);
   if (bthdr->reason)
      return(BTREE_ERR + 1003);
   if (bthdr->blk->count > bthdr->nodesz)
      {
      fprintf(stderr,"btisrt: block count %d "
	 "greater than configuration %d\n",
	 bthdr->blk->count, bthdr->nodesz);
      bthdr->reason = BTREE_ERR + 1004;
      return(BTREE_ERR + 1004);
      } /* invalid count */
   if (bthdr->blk->count == bthdr->nodesz)
      {
      if (ofst == bthdr->rootofst)
         {
         btspltrt(bthdr);
         if (bthdr->reason)
            return(BTREE_ERR + 1005);
         } /* if root */
      else
         {
         btsplit(bthdr);
         if (bthdr->reason)
            return(BTREE_ERR + 1006);
         } /* not root */
      } /* btsplit */
   if (bthdr->blk->count >= bthdr->nodesz)
      {
      fprintf(stderr,"btisrt: block overflow\n");
      bthdr->reason = BTREE_ERR + 1007;
      return(BTREE_ERR + 1007);
      } /* block overflow */
   currcd = 1;
   rcdptr = (unsigned char *) bthdr->blk + bthdr->blkhdr;
   rslt = 999999;
   while (currcd <= bthdr->blk->count)
      {
      if (bthdr->blk->count > bthdr->nodesz)
	 {
         fprintf(stderr,"btisrt: count exceeds nodesz\n");
	 fprintf(stderr,"count %d nodesz %d\n",
	    bthdr->blk->count, bthdr->nodesz);
         bthdr->reason = BTREE_ERR + 1008;
         return(BTREE_ERR + 1008);
	 } /* if invalid count */
      else if (bthdr->blk->count == bthdr->nodesz)
         {
         if (bthdr->tail->prev->ofst
	    == bthdr->rootofst)
            {
            btspltrt(bthdr);
            if (bthdr->reason)
               return(BTREE_ERR + 1009);
            } /* if root */
         else
            {
            btsplit(bthdr);
            if (bthdr->reason)
               return(BTREE_ERR + 1010);
            } /* not root */
         rcdptr = (unsigned char *) bthdr->blk + bthdr->blkhdr;
	 currcd = 1;
         } /* btsplit */
      p = (unsigned char *) rcdptr
	 + bthdr->rcdhdr
	 + bthdr->keyofst;
      q = (unsigned char *) p + bthdr->keylen;
      r = (unsigned char *) rcd + bthdr->keyofst;
      while (p < q)
	 {
	 rslt = *r - *p;
	 if (rslt) break;
	 p++;
	 r++;
	 } /* for each byte in key */
      if (!rslt)
	 {
	 p = (unsigned char *) rcdptr
	    + bthdr->rcdhdr;
	 if (*(p + bthdr->dletofst) == DELETED)
	    {
	    q = (unsigned char *) p + bthdr->rcdlen;
	    r = (unsigned char *) rcd;
	    while (p < q) *p++ = *r++;
            btputblk(bthdr,
	       bthdr->tail->prev->ofst,
	       bthdr->blk);
	    if (bthdr->reason)
               return(BTREE_ERR + 1020);
            btflustk(bthdr);
            if (bthdr->reason)
               return(BTREE_ERR + 1021);
            return(ISRTGOOD);
	    } /* replace deleted index node */
         else
            {
            fprintf(stderr,"btisrt: duplicate key ");
            p = (unsigned char *) rcd + bthdr->keyofst;
            q = (unsigned char *) p + bthdr->keylen;
            while (p < q)
               fprintf(stderr,"%c", *p++);
            fprintf(stderr,"\n");
            fprintf(stderr,"btisrt: duplicate key ");
            p = (unsigned char *) rcd + bthdr->keyofst;
            q = (unsigned char *) p + bthdr->keylen;
            while (p < q)
               fprintf(stderr,"%02x", *p++);
            fprintf(stderr,"\n");
            fprintf(stderr,"btisrt: ofst %d\n",
               bthdr->ofst);
            bthxmem(bthdr->blksz,
               (unsigned char *) bthdr->blk);
            return(ISRTDUP);
            } /* not filling deleted index node */
	 } /* dup rcd key */
      if (rslt > 0
	 && bthdr->blk->right != 0
	 && currcd >= bthdr->blk->count)
	 {
         btgetblk(bthdr,bthdr->blk->right);
         if (bthdr->reason)
            return(BTREE_ERR + 1011);
         btpush(bthdr,bthdr->ofst);
         if (bthdr->reason)
            return(BTREE_ERR + 1012);
         rcdptr = (unsigned char *) bthdr->blk + bthdr->blkhdr;
         currcd = 1;
	 continue;
	 } /* greater than currend node */
      node = (bt_nodefmt *) rcdptr;
      if (rslt < 0
	 && currcd <= bthdr->blk->count
	 && node->left != 0)
	 {
         btgetblk(bthdr,node->left);
         if (bthdr->reason)
            return(BTREE_ERR + 1013);
         btpush(bthdr,bthdr->ofst);
         if (bthdr->reason)
            return(BTREE_ERR + 1014);
         rcdptr = (unsigned char *) bthdr->blk + bthdr->blkhdr;
         currcd = 1;
	 continue;
	 } /* greater than currend node */
      if (rslt < 0) break;
      currcd++;
      rcdptr = (unsigned char *) rcdptr
	 + bthdr->rcdhdr
	 + bthdr->rcdlen;
      } /* for each record in block */
   if (currcd - 1 > bthdr->blk->count)
      {
      fprintf(stderr,"btisrt: invalid block count\n");
      fprintf(stderr,"currcd %d "
	 "blk->count %d\n",
	 currcd, bthdr->blk->count);
      bthdr->reason = BTREE_ERR + 1015;
      return(BTREE_ERR + 1015);
      } /* invalide block count */

   p = (unsigned char *) rcdptr;
   q = (unsigned char *) p
      + ((bthdr->blk->count - currcd + 1)
	 * bthdr->rcdnode);
   r = (unsigned char *) bthdr->blkleft;
   while (p < q) *r++ = *p++;

   p = (unsigned char *) rcdptr;
   q = (unsigned char *) p + bthdr->rcdhdr;
   while (p < q) *p++ = '\0';

   q = (unsigned char *) p + bthdr->rcdlen;
   r = (unsigned char *) rcd;
   while (p < q) *p++ = *r++;

   q = (unsigned char *) p
      + ((bthdr->blk->count - currcd + 1)
	 * bthdr->rcdnode);
   r = (unsigned char *) bthdr->blkleft;
   while (p < q) *p++ = *r++;

   p = (unsigned char *) bthdr->blkleft;
   q = (unsigned char *) p
      + ((bthdr->blk->count - currcd + 1)
	 * bthdr->rcdnode);
   while (p < q) *p++ = '\0';

   bthdr->blk->count++;
   if (bthdr->blk->count > bthdr->nodesz)
      {
      fprintf(stderr,"btisrt: invalid block count\n");
      fprintf(stderr,"blk->count %d\n",
	 bthdr->blk->count);
      fprintf(stderr,"nodesz %d (records per node)\n",
	 bthdr->nodesz);
      bthdr->reason = BTREE_ERR + 1016;
      return(BTREE_ERR + 1016);
      } /* invalid block count */
   btputblk(bthdr,
      bthdr->tail->prev->ofst,
      bthdr->blk);
   if (bthdr->reason)
      return(BTREE_ERR + 1017);

   p = (unsigned char *) bthdr->blk;
   q = (unsigned char *) bthdr->blk + bthdr->blksz;
   while (p < q) *p++ = '\0';
   btflustk(bthdr);
   if (bthdr->reason)
      return(BTREE_ERR + 1018);
   return(ISRTGOOD);
   } /* btisrt */
